from textcrafts import deepRank as dr
from collections import defaultdict

## PARAMS for Dialog Engines in query.py

class pytalk_params(dr.craft_params):
  def __init__(self):
    super().__init__()
    self.corenlp = True
    self.quiet = True
    self.summarize = True
    self.quest_memory = 1
    self.max_answers = 3
    self.repeat_answers = 'yes'
    self.by_rank = 'no'
    self.personalize = 20
    self.show=False

    self.sent_count = 8
    self.word_count = 5
    self.rel_count = 10
    self.dot_count = 20

params=pytalk_params()

def multi_dict():
  return defaultdict(set)

class DialogAgent(dr.GraphMaker):
  def __init__(self,file_name = None, text = None, corenlp = False):
    super().__init__(params = params,
                     file_name = file_name, text = text, corenlp = corenlp)
    self.clear_index()

  def clear_index(self):
    self.source = None
    self.target = None

  def source_target_pair(self):
    for e, k in self.edgesInSent():
      f, ft, r, t, tt = e
      yield f, (k,t, (ft, r, tt))

  def target_source_pair(self):
    for e, k in self.edgesInSent():
      f, ft, r, t, tt = e
      yield t, (k, f, (ft,r,tt))

  def source_index(self):
    m = multi_dict()
    for x, y in self.source_target_pair():
      m[x].add(y)
    return m

  def target_index(self):
    m = multi_dict()
    for x, y in self.target_source_pair():
      m[x].add(y)
    return m

  def index(self):
    self.clear_index()
    if not self.source:
      self.source = self.source_index()
    if not self.target:
      self.target = self.target_index()

  def stats(self):
    return (len(self.source), len(self.target))

  def chat_about(self,question=None):

    #self.load(self.fname)
    pr=self.pagerank()
    #chat(self.index()
    #for x in self.source.items() : print(x,'\n')

    summary=list(self.bestSentences(self.params.sent_count))
    sents={s for s,_ in summary}
    print('SUMMARY',sents)
    dr.print_summary(summary)
    self.index()
    #print('SELF',self.stats())
    db=set(self.source).union(set(self.target))
    #print('DB',db)

    QA = DialogAgent()
    def query(QA, text, echo=False):
      if echo: print('?--', text)
      QA.digest(text)
      pr = QA.pagerank()
      qpr = self.rerank(pers=pr)

      QA.index()
      #print('QA', QA.stats())
      qa=set(QA.source).union(set(QA.target))
      #print('QA', qa)
      shared = { x for x in qa.intersection(db)
                   if dr.maybeWord(x) and not dr.isStopWord(x)
               }
      #print('SHARED',shared)

      good=set()
      for w in shared :
        for (s,k,_) in self.source[w] :
          if dr.isSent(k) : good.add(k)
        for (s, k, _) in self.target[w]:
          if dr.isSent(k) : good.add(k)

      def inhabits(x) : return x in good
      answers = list(self.bestSentencesByRank(self.params.max_answers,filter=inhabits))
      new_answers=[(s,a) for s,a in answers if s not in sents]
      if new_answers : answers=new_answers

      dr.print_summary(answers)
      if self.params.show:
        dr.query_edges_to_dot(QA)
        self.toDot(self.params.dot_count, dr.maybeWord, svo=True, show=self.params.show)

    if isinstance(question, list):
      for q in question:
        query(QA,q,echo=True)
    elif question :
      query(QA,question,echo=True)
    else :
      while (True):
        question = input('?-- ')
        if not question: break
        query(QA,question)


def chat(fname) :
  DialogAgent().chat_about("examples/"+fname+".txt")

def dialog_about(fname,query=None) :
  DialogAgent(file_name=fname+".txt").chat_about(question=query)

def ranked_txt_quest(Folder, FNameNoSuf, QuestFileNoSuf):
  Q = []
  qfname = Folder + "/" + QuestFileNoSuf + ".txt"
  qs = list(file2seq(qfname))
  print('qs',qs)
  dialog_about(Folder+"/"+FNameNoSuf,qs)

def file2seq(fname):
  with open(fname, 'r') as f:
    for l in f: yield l.strip()

def t0():
  dialog_about('examples/tesla',
               "How I have a flat tire repaired?")


def t0a():
  dialog_about('examples/tesla',
               "How I have a flat tire repaired?  \
               Do I have Autopilot enabled? \
               How I navigate to work? Should I check tire pressures?")

def t0b() :
  ranked_txt_quest('examples','tesla','quests')

def t1():
  dialog_about('examples/bfr',
               "What space vehicles SpaceX develops?")

def t2():
  # dialog_about('examples/bfr')
  dialog_about('examples/hindenburg',
               "How did the  fire start on the Hindenburg?")


def t3():
  dialog_about('examples/const',
               # "How many votes are needed for the impeachment of a President?"
               'How can a President be removed from office?'
               )


def t4():
  dialog_about('examples/summary',
               "How we obtain summaries and keywords from dependency graphs?")


def t5():
  dialog_about('examples/heaven',
               "What does the Pope think about heaven?")


def t6():
  dialog_about('examples/einstein',
               "What does quantum theory tell us about our \
                description of reality for an observer?")


def t7():
  dialog_about('examples/kafka',
               # "What does the doorkeeper say about entering?"
               "Why does K. want access to the law at any price?"
               )


def t8():
  dialog_about('examples/test',
               "Does Mary have a book?")


def t9():
  dialog_about('examples/relativity',
               "What happens to light in the presence of gravitational fields?")


def t11() :
  ranked_txt_quest('examples','texas','texas_quest')

def t12() :
  ranked_txt_quest('examples','geo','geo_quest')

def qgo() :
  t1()
  t2()
  t3()
  t4()
  t5()
  t6()
  t7()
  t8()
  t9()
  t0()

if __name__ == '__main__'  :
  pass
  t11()





from collections import defaultdict
import subprocess
from pathlib import Path
import math
import networkx as nx

from .nlp import *
from .vis import pshow,gshow

from nltk.corpus import stopwords

stop_words=set(stopwords.words('english')).union({'|(){}[]%'})
client = NLPclient()


def run_with(fname,query=True,show=show) :
  '''
  Activates dialog about document in <fname>.txt with questions
  in <fname>_quests.txt
  Assumes stanford corenlp server listening on port 9000
  with annotators listed in params.py  available.
  '''
  t = Talker(from_file=fname+'.txt')
  t.show_summary()
  t.show_keywords()
  if show :
    pshow(t,file_name=fname)
  if query:
    t.query_with(fname+'_quest.txt')
    pshow(t,file_name=fname+"_quest",show=show)

def tprint(*args) :
  if trace : print(*args)

def say(what) :
  print(what)
  if not quiet : subprocess.run(["say", what])

def tload(infile) :
  tprint('LOADING:',infile,'\n')
  with open(infile, 'r') as f: text = f.read()
  return digest(text)

def jload(infile) :
  with open(infile, 'r') as f:
    res = json.load(f)
    return res

def jsave(infile,outfile):
  d=tload(infile)
  with open(outfile,'w') as g:
    json.dump(d,g,indent=0)

def exists_file(fname) :
  path = Path(fname)
  return path.is_file()

def load(fname) :
  if fname[-4:]==".txt":
    if force:
      db = tload(fname)
    else :
      jfname=fname[:-4]+".json"
      if not exists_file(jfname) :
        jsave(fname,jfname)
      db=jload(jfname)
  else:
    db = jload(fname)
  return db

def get_quests(qs) :
  if not isinstance(qs,list) :
    qfname=qs
    with open(qfname,'r') as f:
      qs = list(l.strip() for l in f)
  return qs

def digest(text) :
  l2occ = defaultdict(list)
  sent_data=[]
  for i,xss in enumerate(client.extract(text)) :
    lexs,deps,ies=xss
    sent,lemma,tag,ner=[],[],[],[]
    for j,t in enumerate(lexs):
      w,l,p,n=t
      wi=len(l2occ)
      l2occ[l].append((i,j))
      sent.append(w)
      lemma.append(l)
      tag.append(p)
      ner.append(n)
    d=(tuple(sent),tuple(lemma),tuple(tag),
       tuple(ner),tuple(deps),tuple(ies))
    sent_data.append(d)
  #tprint('DIGESTED')
  return sent_data,l2occ

SENT,LEMMA,TAG,NER,DEP,IE=0,1,2,3,4,5

def rel_from(d):
  def to_lems(ux):
    f,t=ux
    if f>=0:
      for u in range(*ux):
        yield lemma[u],tag[u]
  rs,svos=set(),set()
  for ts in d[IE] :
    for t in ts :
      sx, vx, ox = t
      lemma = d[LEMMA]
      tag=d[TAG]
      sub = tuple(to_lems(sx))
      rel = tuple(to_lems(vx))
      ob = tuple(to_lems(ox))
      res = (sub, rel, ob)
      s=()
      for l,tl in sub:
        if tl[0]=='N' :
          s=l
      o=()
      for l, tl in ob:
        if tl[0] == 'N':
          o = l
      v=()
      for l, tl in rel:
        if tl[0] == 'V':
          v = l
      rs.add(res)
      svo=s,v,o
      if () in svo : continue
      svos.add(svo)

  return tuple(rs),tuple(svos)

def dep_from(id,d):
  deps=d[DEP]
  lemmas=d[LEMMA]
  tags=d[TAG]
  for dep in deps :
    f, r, t = dep
    if t == -1 : target,ttag=id,'SENT'
    else: target,ttag = lemmas[t],tags[t]
    res = lemmas[f],tags[f],r,target,ttag
    yield res

def deps_from(id,d) :
  return tuple(t for t in dep_from(id,d))

def comp_from(id,d) :
  for x in dep_from(id,d) :
    f,tf,rel,t,tt=x
    #ppp(f,t)
    if (rel == 'compound' or rel == 'amod') and \
       good_word(f) and good_word(t) and \
       good_tag(tf) and good_tag(tt) :
      yield (f,t)

def comps_from(id,d) :
  return tuple(t for t in comp_from(id,d) if t)

def to_edges(db) :
  sent_data,l2occ=db
  for id,sd in enumerate(sent_data) :
    for dep in dep_from(id,sd):
      f,f_,r,t,t_=dep
      if r == 'punct': continue
      elif r in ['nsubj','dobj','iobj'] or t_[0]=='V':
        yield (id, f) # sent to predicate
        yield (t,f) # pred to arg
        yield (f,id) # arg to sent
      elif f in stop_words or t in stop_words:
        continue
      elif r=='ROOT' :
        yield (t,f)
      else :
        yield (f,t)
    if compounds :
      for ft in comps_from(id, sd):
          f,t=ft
          yield f, ft #parts to compound
          yield t, ft
          yield ft,id # compound to sent
          yield ft,ft # to self


def get_avg_len(db) :
  sent_data,_=db
  lens=[len(x[LEMMA]) for x in sent_data]
  n=len(lens)
  s=sum(lens)
  return round(s/n)

def to_graph(db,personalization=None) :
  g = nx.DiGraph()
  for e in to_edges(db) :
    f,t=e
    g.add_edge(f,t)
  try :
     pr=nx.pagerank(g,personalization=personalization)
  except :
    n=g.number_of_nodes()
    pr=dict()
    for l in db[1] : pr[l]=1/n
  return g,pr

def rank_sort(pr) :
  by_rank=[(x,r) for (x,r) in pr.items()]
  by_rank.sort(key=lambda x : x[1],reverse=True)
  return by_rank


def ners_from(d):
  ners=[]
  for j, ner in enumerate(d[NER]):
    lemma = d[LEMMA][j]
    if ner != 'O': ners.append((lemma,ner))
  return tuple(ners)

def show_db(db) :
    sent_data,l2occ = db
    for i,d in enumerate(sent_data) :
      for trip in rel_from(d): print('TRIPLES:', trip)
      for dep in deps_from(i,d) : print('DEPENDS:',dep)
      print("NERS",ners_from(d))
    print('')

def materialize(db) :
  sent_data,l2occ= db
  for i,d in enumerate(sent_data) :
      rels,svos = rel_from(d)
      deps=deps_from(i,d)
      comps=comps_from(i,d) # or directly from deps
      ners=ners_from(d)
      yield tuple(d[SENT]),tuple(d[LEMMA]),tuple(d[TAG]),\
            ners,rels,svos,deps,comps

def svos(fname) :
  t=Talker(from_file=fname)
  db=t.db
  for i,m in enumerate(materialize(db)) :
    lemmas,words,tags,ners,rels,svos,deps = m
    for s,v,o in svos :
      yield s,v,o,i

def answer_quest(q,talker) :
    db=talker.db
    sent_data,l2occ=db
    matches = defaultdict(set)
    answerer=Talker(from_text=q)
    q_sent_data,q_l2occ=answerer.db
    #gshow(answerer.g)
    unknowns=[]
    for q_lemma in q_sent_data[0][LEMMA]:
       if q_lemma in stop_words or q_lemma in ".?" : continue
       ys = l2occ.get(q_lemma)
       if not ys :
         unknowns.append(q_lemma)
         continue
       for sent,pos in ys:
         matches[sent].add(q_lemma)
         #else : print('UNMATCHED LEMMA',q_lemma,q_tag,tag)
    if unknowns: tprint("UNKNOWNS:", unknowns,'\n')
    best=[]
    if pers :
      talker.pr=nx.pagerank(talker.g,personalization=answerer.pr)
    for (id, shared) in matches.items() :
      sent=sent_data[id][SENT]
      r=answer_rank(id,shared,sent,talker)
      best.append((r,id,shared,sent))
    best.sort(reverse=True)

    answers=[]
    for i,b in enumerate(best):
      if i >= max_answers : break
      rank, id, shared, sent = b
      answers.append((id,sent,round(rank,3),shared))
    if not answers_by_rank : answers.sort()
    return answers

def answer_rank(id,shared,sent,talker) :
  lshared = len(shared)
  lsent = len(sent)
  lavg=talker.avg_len
  srank=talker.pr.get(id)
  def get_freq(x) :
    return len(talker.db[1].get(x))

  freq=sum(get_freq(x) for x in shared)
  if not srank :
    srank=0
  good=lshared*math.exp(srank)
  if lsent>2*lavg :
    long=math.sqrt(lsent)
  else :
    long=1
  freq=math.log(freq)
  #ppp('HOW:', good, long, freq, long * freq, shared)
  r=good/(1+long*freq)

  r=math.tanh(r)
  return r

def answer_with(talker,qs)     :
  qs = get_quests(qs)
  if qs:
    for q in qs :
      if not q :break
      interact(q,talker)
  else:
    while True:
      q=input('> ')
      if not q : break
      interact(q,talker)

def interact(q,talker):
  tprint('----- QUERY ----\n')
  say(q)
  print('')
  for info, sent, rank, shared in answer_quest(q, talker):
    print(info,end=': ')
    say(nice(sent))
    tprint('  ', shared, rank)
  print('')
  tprint('------END-------', '\n')

class Talker :
  def __init__(self,from_file=None,from_text=None,
               sk=sum_count,wk=key_count,show=show):
    if from_file:
       self.db=load(from_file)
       self.from_file=from_file
    elif from_text :
       self.db=digest(from_text)
    else :
      assert from_file or from_text
    self.sum_count=sk
    self.key_count=wk
    self.avg_len = get_avg_len(self.db)
    self.g,self.pr=to_graph(self.db)
    #self.get_sum_and_words(sk,wk)
    self.summary, self.keywords = self.extract_content(sk, wk)

  def query_with(self,qs):
    qs = get_quests(qs)
    answer_with(self,qs)

  def get_tagged(self,w):
    l2occ=self.db[1]
    sent_data=self.db[0]
    occs=l2occ.get(w)
    tags=set()
    words=set()
    for i,j in occs:
      word = sent_data[i][SENT][j]
      tag=sent_data[i][TAG][j]
      words.add(word)
      tags.add(tag)
    return words,tags

  def extract_content(self,sk,wk):
    def good_sent(ws) :
      return len(ws)<=self.avg_len+2
    def nice_word(x,good_tags='N') :
      ws, tags = self.get_tagged(x)
      ncount = 0
      for tag in tags:
        if tag[0] in good_tags:
          ncount += 1
      if ncount > len(tags) // 2:
        cx = x.capitalize()
        if cx in ws: x = cx
        return x
      else:
        return None

    sents,words=[],[]
    by_rank=rank_sort(self.pr)
    for i  in range(len(by_rank)):
      x,r=by_rank[i]
      if sk and isinstance(x,int) :
        ws=self.db[0][x][SENT]
        if good_sent(ws) :
          sk-=1
          sents.append((x,ws))
      elif wk and good_word(x) :
        x=nice_word(x)
        if x:
          wk -= 1
          words.append(x)
      elif wk and isinstance(x,tuple) :
          x=tuple(map(nice_word,x))
          if all(x) :
            wk -= 1
            words.append(x)
    sents.sort(key=lambda x: x[0])
    summary=[(s,nice(ws)) for (s,ws) in sents]
    self.by_rank=by_rank # to be used when needed
    return summary,words

  def show_summary(self):
    say('SUMMARY:')
    for s in self.summary:
      say(s)
    print('')

  def show_keywords(self):
    print('KEYWORDS:')
    print(self.keywords)
    print('')


# helpers
def nice(ws) :
  ws=[cleaned(w) for w in ws]
  sent=" ".join(ws)
  #print(sent)
  sent=sent.replace(" 's","'s")
  sent=sent.replace(" ,",",")
  sent=sent.replace(" .",".")
  sent = sent.replace('``', '"')
  sent = sent.replace("''", '"')
  return sent

def good_word(w) :
  return isinstance(w,str) and w.isalpha() and w not in stop_words

def good_tag(tag,starts="NVJA"):
  c=tag[0]
  return c in starts



from .params import *
from graphviz import Digraph as DotGraph
from wordcloud import WordCloud
import matplotlib.pyplot as plt


def showGraph(dot, show=True, file_name='textgraph.gv'):
  dot.render(file_name, view=show)

def gshow(g, file_name='temp.gv', show=1):
  dot = DotGraph()
  for e in g.edges():
    f, t = e
    # w = g[f][t]['weight']
    w = ''
    dot.edge(str(f), str(t), label=str(w))
  dot.render(file_name, view=show>1)

def pshow(t, k=24,file_name="temp",show=show):
  def t2s(x) :
    if isinstance(x,tuple) :
      return " ".join(x)
    return x
  sum, kws = t.extract_content(5, k)
  #ppp(t.by_rank)
  d=dict()
  s=set()
  for kw in kws:
    if isinstance(kw,tuple) :
      kw0=tuple(map(lambda x: x.lower(),kw))
      d[t2s(kw)]=t.pr[kw0]
      s.add(kw0)
    else :
      d[kw]=t.pr[kw.lower()]
      s.add(kw)
  #ppp("DDDD",d)
  show_ranks(d,file_name=file_name+"_cloud.pdf",show=show)
  topg=t.g.subgraph(s)
  gshow(topg,file_name+".gv",show=show)

def show_ranks(rank_dict,file_name="cloud.pdf",show=show) :
  cloud=WordCloud(width=800,height=400)
  cloud.fit_words(rank_dict)
  f=plt.figure()
  plt.imshow(cloud, interpolation='bilinear')
  plt.axis("off")
  if show>1 : plt.show()
  f.savefig(file_name,bbox_inches='tight')
  plt.close('all')

if __name__=="__main__":
  d = {'a': 0.1, 'b': 0.2, 'c': 0.33, 'd': 0.2}
  show_ranks(d)



annotators=['tokenize','ssplit','pos','lemma','depparse','ner']+\
           ['natlog','openie']
openie='openie' in annotators

sum_count=4
key_count=10
quiet=True
max_answers=3
answers_by_rank=True
trace=1
force=False
from nltk.stem import PorterStemmer
#stemmer=PorterStemmer()
stemmer=None
lower=True
pers=True
compounds=True
show=1
def ppp(*args) :
  print('DEBUG:',end='')
  print(*args)
